-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

local max = math.max
local min = math.min

local function updateGFX(storage, dt)
  storage.remainingVolume = storage.initialStoredEnergy > 0 and storage.storedEnergy * storage.itemEnergy or 0
  storage.remainingRatio = storage.initialStoredEnergy > 0 and storage.storedEnergy / storage.initialStoredEnergy or 0

  for k, v in pairs(storage.fuelNodes) do
    obj:setNodeMass(k, v + storage.storedEnergy * storage.fuelNodeMassCoef)
  end

  if storage.currentLeakRate > 0 then
    storage:setRemainingVolume(storage.remainingVolume - storage.currentLeakRate * dt)
  end
end

local function setRemainingVolume(storage, volume)
  storage.storedEnergy = min(max(volume, 0), storage.capacity) * storage.itemEnergy
end

local function setRemainingRatio(storage, ratio)
  storage.storedEnergy = storage.initialStoredEnergy * min(max(ratio, 0), 1)
end

local function onBreak(storage)
  storage.currentLeakRate = storage.brokenLeakRate
end

local function registerDevice(storage, device)
  storage.assignedDevices[device.name] = device
end

local function reset(storage)
  storage.currentLeakRate = 0
  storage.storedEnergy = storage.startingStackSize * storage.itemEnergy

  --apply final weight as soon as possible
  for k, v in pairs(storage.fuelNodes) do
    obj:setNodeMass(k, v + storage.storedEnergy * storage.fuelNodeMassCoef)
  end

  storage.remainingRatio = storage.initialStoredEnergy > 0 and storage.storedEnergy / storage.initialStoredEnergy or 0
end

local function deserialize(storage, data)
  if data.remainingVolume then
    storage:setRemainingVolume(data.remainingVolume)
  end
end

local function serialize(storage)
  return {remainingVolume = storage.remainingVolume}
end

local function new(jbeamData)
  local storage = {
    name = jbeamData.name,
    type = jbeamData.type,
    itemType = jbeamData.energyType or "mc_008064",
    assignedDevices = {},
    remainingRatio = 1,
    remainingVolume = 1,
    currentLeakRate = 0,
    breakTriggerBeam = jbeamData.breakTriggerBeam,
    jbeamData = jbeamData,
    updateGFX = updateGFX,
    setRemainingVolume = setRemainingVolume,
    setRemainingRatio = setRemainingRatio,
    deserialize = deserialize,
    serialize = serialize,
    registerDevice = registerDevice,
    onBreak = onBreak,
    reset = reset
  }
  
 --energy = operations per fuel taken from Java edition * million * 5
  if storage.itemType == "mc_100001" then
    storage.itemEnergy = 100 * 1000000 * 5
	storage.stackSize = 1
  elseif storage.itemType == "mc_080064" then
    storage.itemEnergy = 80 * 1000000 * 5
	storage.stackSize = 64
  elseif storage.itemType == "mc_020064" then
    storage.itemEnergy = 20 * 1000000 * 5
	storage.stackSize = 64
  elseif storage.itemType == "mc_012064" then
    storage.itemEnergy = 12 * 1000000 * 5
	storage.stackSize = 64
  elseif storage.itemType == "mc_008064" then
    storage.itemEnergy = 8 * 1000000 * 5
	storage.stackSize = 64
  elseif storage.itemType == "mc_002001" then
    storage.itemEnergy = 2 * 1000000 * 5
	storage.stackSize = 1
  elseif storage.itemType == "mc_001564" then
    storage.itemEnergy = 1.5 * 1000000 * 5
	storage.stackSize = 64
  elseif storage.itemType == "mc_001516" then
    storage.itemEnergy = 1.5 * 1000000 * 5
	storage.stackSize = 16
  elseif storage.itemType == "mc_001501" then
    storage.itemEnergy = 1.5 * 1000000 * 5
	storage.stackSize = 1
  elseif storage.itemType == "mc_001064" then
    storage.itemEnergy = 1 * 1000000 * 5
	storage.stackSize = 64
  elseif storage.itemType == "mc_001016" then
    storage.itemEnergy = 1 * 1000000 * 5
	storage.stackSize = 16
  elseif storage.itemType == "mc_001001" then
    storage.itemEnergy = 1 * 1000000 * 5
	storage.stackSize = 1
  elseif storage.itemType == "mc_000764" then
    storage.itemEnergy = 0.75 * 1000000 * 5
	storage.stackSize = 64
  elseif storage.itemType == "mc_000564" then
    storage.itemEnergy = 0.5 * 1000000 * 5
	storage.stackSize = 64
  elseif storage.itemType == "mc_000364" then
    storage.itemEnergy = 0.335 * 1000000 * 5
	storage.stackSize = 64
  elseif storage.itemType == "mc_000264" then
    storage.itemEnergy = 0.25 * 1000000 * 5
	storage.stackSize = 64
  end
  storage.itemEnergy = storage.itemEnergy or 0 --MJ/kg

  storage.capacity = storage.stackSize
  if jbeamData.startingStackSize == nil then
    storage.startingStackSize = storage.stackSize
  else
    storage.startingStackSize = min(jbeamData.startingStackSize, storage.stackSize)
  end
  storage.storedEnergy = storage.startingStackSize * storage.itemEnergy

  storage.brokenLeakRate = 2 --lose 2 fuel items / second

  storage.fuelNodeMassCoef = 0
  storage.fuelNodes = {}
  local fuelNodeCount = 0
  if jbeamData.fuel and jbeamData.fuel._engineGroup_nodes then
    for _, n in pairs(jbeamData.fuel._engineGroup_nodes) do
      storage.fuelNodes[n] = v.data.nodes[n].nodeWeight --save initial mass as the offset for the fuel node weights
      fuelNodeCount = fuelNodeCount + 1
    end
    if fuelNodeCount > 0 and storage.itemEnergy > 0 then
      storage.fuelNodeMassCoef = 1 / (storage.itemEnergy * fuelNodeCount) --calculate weight per energy left
    end
  end

  --apply final weight as soon as possible
  for k, v in pairs(storage.fuelNodes) do
    obj:setNodeMass(k, v + storage.storedEnergy * storage.fuelNodeMassCoef)
  end

  storage.initialStoredEnergy = storage.capacity * storage.itemEnergy
  storage.remainingRatio = storage.initialStoredEnergy > 0 and storage.storedEnergy / storage.initialStoredEnergy or 0

  storage.jbeamData = jbeamData

  return storage
end

M.new = new

return M
